@@ -1,8 +1,9 @@
 from functools import partial, wraps
 from collections import deque
-from typing import Union, Callable, Iterable, Any, Optional
+from typing import Any
 from dataclasses import dataclass
 
+from i2.signatures import Sig, Parameter
 from lined.util import func_name, partial_plus, n_required_args
 from lined.base import Line
 
@@ -15,6 +16,13 @@ def identity(x):
 def blind(x, output):
     """Takes one argument, and returns it as is.
     The output is meant to be bound by currying (functools.partial)
+
+    >>> true_no_matter_what = partial(blind, output=True)
+    >>> false_no_matter_what = partial(blind, output=False)
+    >>> true_no_matter_what(42)
+    True
+    >>> false_no_matter_what(42)
+    False
     """
     return output
 
@@ -25,7 +33,9 @@ false_no_matter_what = partial(blind, output=False)
 
 def _extract_first_argument(args: tuple, kwargs: dict):
     """
-    Returns the tuple (X, _args, _kwargs) where X is the first argument (found in either args or kwargs), and _args, _kwargs are the same (with X removed)
+    Returns the tuple (X, _args, _kwargs) where X is the first argument (
+    found in either args or kwargs), and _args, _kwargs are the same (with X
+    removed)
 
     >>> _extract_first_argument((1,2,3), {'d': 4})
     (1, [2, 3], {'d': 4})
@@ -46,7 +56,7 @@ def _extract_first_argument(args: tuple, kwargs: dict):
         return first_arg_val, [], kwargs
 
 
-########################################################################################################################
+# ------------------------------------------------------------------------------
 
 from operator import le
 from typing import Union, Callable, Generator, Iterable
@@ -58,13 +68,19 @@ def if_then_else(
 ):
     """Implement the if-then-else logic as a function.
 
-    >>> if_then_else('world', if_func=lambda x: x == 'world', then_func="hello {}".format, else_func=lambda x: x * 2)
+    >>> if_then_else(
+    ...     'world',
+    ...     if_func=lambda x: x == 'world',
+    ...     then_func="hello {}".format,
+    ...     else_func=lambda x: x * 2)
     'hello world'
-    >>> if_then_else('bora', if_func=lambda x: x == 'world', then_func="hello {}".format, else_func=lambda x: x * 2)
+    >>> if_then_else('bora', if_func=lambda x: x == 'world',
+    ... then_func="hello{}".format, else_func=lambda x: x * 2)
     'borabora'
 
     Really, it's meant to be curried to make functional components.
-    For example, to make a function that ensures that a string is encapsulated in a tuple, we could do this:
+    For example, to make a function that ensures that a string is
+    encapsulated in a tuple, we could do this:
 
     >>> def is_a_str(x): return isinstance(x, str)
     >>> def make_it_a_tuple(x): return tuple([x])
@@ -107,10 +123,25 @@ cast_to_tuple_if_non_iterable_or_a_string = partial(
 )
 
 
-########################################################################################################################
+# ------------ Tools for iterables ---------------------------------------------
+
+
+from itertools import groupby
 
 
 class Command:
+    """Make a no-input callable that will execute a specific function call.
+
+    >>> command = Command(sum, [1, 2, 3])
+    >>> command()
+    6
+    >>> command = Command(print, 'hello', 'world', sep=', ')
+    >>> command()
+    hello, world
+
+    See: https://en.wikipedia.org/wiki/Command_pattern
+    """
+
     def __init__(self, func, *args, **kwargs):
         self.func, self.args, self.kwargs = func, args, kwargs
 
@@ -119,7 +150,8 @@ class Command:
 
 
 class ItemsNotSorted(RuntimeError):
-    """Use to indicate that two consecutive items where not in the expected order"""
+    """Use to indicate that two consecutive items where not in the expected
+    order"""
 
 
 def raise_(exception):
@@ -128,6 +160,40 @@ def raise_(exception):
 
 raise_not_sorted_error = Command(raise_, ItemsNotSorted)
 
+from itertools import groupby
+
+
+def enumerate_groups(iterable, key=None, start=0):
+    """Get enumeration of groups during a groupby call.
+
+    :param iterable: An iterable
+    :param key: The key to use in the groupby logic
+    :param start: Where to start the enumeration (default is 0)
+    :return: A generator of (group_idx, group, item) triples
+
+    >>> iterable = [0, 0, 0, 2, 5, 7, 8, 0, 0, 9, 3, 1]
+    >>> assert list(enumerate_groups(iterable, key=lambda x: x > 0)) == [
+    ...  (0, False, 0),
+    ...  (0, False, 0),
+    ...  (0, False, 0),
+    ...  (1, True, 2),
+    ...  (1, True, 5),
+    ...  (1, True, 7),
+    ...  (1, True, 8),
+    ...  (2, False, 0),
+    ...  (2, False, 0),
+    ...  (3, True, 9),
+    ...  (3, True, 3),
+    ...  (3, True, 1),
+    ...  ]
+
+
+    """
+    g = groupby(iterable, key)
+    for group_idx, (group, grouped_items) in enumerate(g, start):
+        for item in grouped_items:
+            yield group_idx, group, item
+
 
 def pairwise(iterable):
     """Yield sliding window pairs
@@ -139,14 +205,16 @@ def pairwise(iterable):
 
 
 def raise_exception(exception: Union[Callable, BaseException], *args, **kwargs):
-    """Raise an exception (from an exception instance, or a callable that makes one"""
+    """Raise an exception (from an exception instance, or a callable that
+    makes one"""
     if isinstance(exception, Callable):
         exception = exception(*args, **kwargs)
     raise exception
 
 
 def consume_until_error(iterable, caught_errors=(Exception,)):
-    """Iterable that will simply exit with out error if one of the caught errors occurs.
+    """Iterable that will simply exit with out error if one of the caught
+    errors occurs.
 
     >>> list(consume_until_error(map(lambda x: 1 / x, [4, 2, 1, 0, -1])))
     [0.25, 0.5, 1.0]
@@ -170,7 +238,9 @@ def _validated_comparison_func(key: Callable):
             return key(x) <= key(y)
 
         return comp_func
-    assert n_required == 2, f"key should be a callable with 1 or 2 required arguments"
+    assert n_required == 2, (
+        f"key should be a callable with 1 or 2 required " f"arguments"
+    )
     return key
 
 
@@ -179,13 +249,16 @@ def check_sorted_during_iteration(
     key: Callable[[Any, Any], bool] = le,
     not_sorted_callback: Union[Callable, BaseException] = raise_not_sorted_error,
 ) -> Generator:
-    """Wrap an iterable so that ordering of the elements is checked at runtime.
+    r"""Wrap an iterable so that ordering of the elements is checked at runtime.
 
     :param iterable: Iterable to consume
     :param key: The function that defines what it means to be sorted.
-        Could be a Any->bool function, which will act like the key argument of builtin sorted for example.
-        Could also be an explicit (element, next_element)->bool function that returns True iff in the right order
-    :param not_sorted_callback: The function to call when two consecutive elements are not sorted. For example:
+        Could be a Any->bool function, which will act like the key argument
+        of builtin sorted for example.
+        Could also be an explicit (element, next_element)->bool function that
+        returns True iff in the right order
+    :param not_sorted_callback: The function to call when two consecutive
+    elements are not sorted. For example:
         - raising an error (the default)
         - logging the information, and skiping the offending element (or not)
     :return: A generator consuming the input iterable
@@ -194,24 +267,30 @@ def check_sorted_during_iteration(
     [1, 2, 3, 4]
 
     >>> try:
-    ...     for i, x in enumerate(check_sorted_during_iteration([2, 4, 3, 6]), 1):
+    ...     for i, x in enumerate(
+    ...             check_sorted_during_iteration([2, 4, 3, 6]), 1):
     ...         print(x)
     ... except ItemsNotSorted:
-    ...     print(f"An ItemsNotSorted exception was raised right after the {i} element (whose value was {x})")
-    ...     print("----> Normally, here, you'd put some thing to actually handle the exception...")
+    ...     print(
+    ...         f"ItemsNotSorted after {i} element (whose value was {x})")
+    ...     print(
+    ...         "----> Normally, here, you'd put exception handling code")
+    ...
     2
     4
-    An ItemsNotSorted exception was raised right after the 2 element (whose value was 4)
-    ----> Normally, here, you'd put some thing to actually handle the exception...
+    ItemsNotSorted after 2 element (whose value was 4)
+    ----> Normally, here, you'd put exception handling code
 
     Now, mind you, you have total control over what sorted means.
     For example, to define it as strict
-    >>> comp = lambda x, y: x > y  # Really, we suggest to use operator.gt here (and other operator module functions!)
+
+    >>> comp = lambda x, y: x > y  # in real life, use operator.gt
     >>> list(check_sorted_during_iteration(iter([4, 3, 2, 1]), key=comp))
     [4, 3, 2, 1]
 
     Now for a more complex example.
-    First we'll define a function that will consume the iterable until an error occurs, returning the elements consumed.
+    First we'll define a function that will consume the iterable until an
+    error occurs, returning the elements consumed.
 
     >>> from lined.tools import consume_until_error
     >>> consume = Line(check_sorted_during_iteration, consume_until_error, list)
@@ -221,7 +300,7 @@ def check_sorted_during_iteration(
     ['a', 'ba', 'cba']
     >>> consume(iterable, len)  # compare based on the length
     ['a', 'ba', 'cba', 'cba', 'back', 'bacca']
-    >>> consume(iterable, lambda x: x[0])  # compare based on the first letter only
+    >>> consume(iterable, lambda x: x[0])  # compare based on first letter only
     ['a', 'ba', 'cba', 'cba']
     >>> # compare based on whether the previous element is a subset of the next:
     >>> consume(iterable, lambda x, y: set(x).issubset(y))
@@ -230,10 +309,14 @@ def check_sorted_during_iteration(
     """
 
     # TODO: Optimization opportunity:
-    #   In this implementation, the key of an element is computed twice (once when element, once when next_element)
-    # TODO: key could be generalized to being a Callable[[element, next_element], bool].
-    #   Though note that it's only an interface flexibility since same could (?) be acheived with a key returning an
-    #   instance of a class such that class.__le__(element, next_element) is what is desired
+    #   In this implementation, the key of an element is computed twice (once
+    #   when element, once when next_element)
+    # TODO: key could be generalized to being a Callable[[element,
+    #  next_element], bool].
+    #   Though note that it's only an interface flexibility since same could
+    #   (?) be acheived with a key returning an
+    #   instance of a class such that class.__le__(element, next_element) is
+    #   what is desired
     key = _validated_comparison_func(key)
     for element, next_element in pairwise(iterable):
         yield element
@@ -242,7 +325,7 @@ def check_sorted_during_iteration(
     yield next_element
 
 
-########################################################################################################################
+# ------------------------------------------------------------------------------
 
 
 def del_fields(d, fields):
@@ -269,6 +352,8 @@ def add_name(obj, name=None):
 
 
 def keys_extractor(keys):
+    """Deprecated: Use operator.itemgetter(*keys) instead."""
+
     def extract(x):
         return tuple((x[i] for i in keys))
 
@@ -276,7 +361,8 @@ def keys_extractor(keys):
 
 
 def apply_to_single_item(func: Callable, item_idx: int):
-    """Get a version of func that applies itself to only the item_idx-th element of the input,
+    """Get a version of func that applies itself to only the item_idx-th
+    element of the input,
     leaving the rest untouched.
 
     That is, apply_to_single_item(func, 2), for example, is a new_func such that
@@ -284,11 +370,13 @@ def apply_to_single_item(func: Callable, item_idx: int):
         new_func([a, b, c, d, e]) == [a, b, func(c), d, e]
     ```
 
-    :param func: A function to apply to a single element of an iterable (that has a [...])
+    :param func: A function to apply to a single element of an iterable (that
+    has a [...])
     :param item_idx: The particular item index to apply function to
     :return:
 
-    >>> apply_to_second_item = apply_to_single_item(func=lambda x: x * 10, item_idx=1)
+    >>> apply_to_second_item = apply_to_single_item(
+    ...     func=lambda x: x * 10, item_idx=1)
     >>> apply_to_second_item([1, 2, 3, 4])
     (1, 20, 2, 3, 4)
     """
@@ -324,9 +412,12 @@ def iterate(iterable: Iterable):
     hello 4
     hello 6
 
-    It could be a bit awkward to have to "consume" the iterable to have it take effect.
-    Just calling  ``pipe([1, 2, 3])`` to get those prints seems like a more natural way.
-    This is where you can use `iterate`. It basically "launches" that consuming loop for you.
+    It could be a bit awkward to have to "consume" the iterable to have it
+    take effect.
+    Just calling  ``pipe([1, 2, 3])`` to get those prints seems like a more
+    natural way.
+    This is where you can use `iterate`. It basically "launches" that
+    consuming loop for you.
 
     >>> pipe = Pipeline(iterize(lambda x: x * 2),
     ...                iterize(lambda x: print(f"hello {x}")),
@@ -343,9 +434,53 @@ def iterate(iterable: Iterable):
         pass
 
 
+def append_output_to_input(func, appender=lambda x, output: (x, output)):
+    """Decorator that makes the function into a function returning its input with output
+
+    ┌─────────────┐
+    │    input    │
+    └─────────────┘
+         │
+         ▼
+    ┌─────────────┐
+    │    func     │
+    └─────────────┘
+         │
+         ▼
+    ┌─────────────────┐
+    │ (input, output) │
+    └─────────────────┘
+
+    >>> func = lambda x: f"hello {x}"
+    >>> func('world')
+    'hello world'
+    >>> new_func = append_output_to_input(func)
+    >>> new_func('world')
+    ('world', 'hello world')
+    """
+    assert n_required_args(func) == 1
+
+    @wraps(func)
+    def _func(x):
+        output = func(x)
+        return appender(x, output)
+
+    return _func
+
+
 def side_call(x, callback):
-    """Identity function that calls a callaback function before returning the input as is
-    (if callback didn't change it)
+    """Identity function that calls a callaback function before returning the
+    input as is (unless the input is mutable and the callback changes it).
+
+    >>> add2 = lambda x: x + 2
+    >>> add2(40)
+    42
+    >>> from functools import partial
+    >>> logger = partial(side_call, callback=lambda x: print(f"input is {x}"))
+    >>> logged_add2 = Line(logger, add2)
+    >>> logged_add2(40)
+    input is 40
+    42
     """
     callback(x)
     return x
@@ -359,7 +494,8 @@ print_and_pass_on = partial_plus(
 )
 
 
-# Function transformers ###################################################################
+# Function transformers
+# ###################################################################
 
 
 def extra_wraps(func, name=None, doc_prefix=""):
@@ -376,8 +512,10 @@ def mywraps(func, name=None, doc_prefix=""):
 
 
 def tail_io(func):
-    """Will apply function only to the tail of tuple inputs, still passing the header on.
-    That is, from a ``x -> func(x)`` function, you get a ``(*header, x) -> (*header, func(x))`` function.
+    """Will apply function only to the tail of tuple inputs, still passing
+    the header on.
+    That is, from a ``x -> func(x)`` function, you get a ``(*header, x) -> (
+    *header, func(x))`` function.
 
     >>> def foo(x):
     ...    return x * 2
@@ -387,7 +525,8 @@ def tail_io(func):
     >>> new_foo = tail_io(foo)
     >>> new_foo((7, 'boo'))
     (7, 'booboo')
-    >>> new_foo(('all', 'items', 'but', 'the', 'last', 'are', 'just', 'passed', 'on', 'boo'))
+    >>> new_foo(('all', 'items', 'but', 'the', 'last', 'are', 'just',
+    ...          'passed', 'on', 'boo'))
     ('all', 'items', 'but', 'the', 'last', 'are', 'just', 'passed', 'on', 'booboo')
 
     """
@@ -403,7 +542,8 @@ def tail_io(func):
 
 def iterize(func, name=None):
     """From an Input->Ouput function, makes a Iterator[Input]->Itertor[Output]
-    Some call this "vectorization", but it's not really a vector, but an iterable, thus the name.
+    Some call this "vectorization", but it's not really a vector, but an
+    iterable, thus the name.
 
     >>> f = lambda x: x * 10
     >>> f(2)
@@ -421,7 +561,8 @@ def iterize(func, name=None):
     >>> pipe(1)
     'hello 2'
 
-    But what if you wanted to use the pipeline on a "stream" of data. The following wouldn't work:
+    But what if you wanted to use the pipeline on a "stream" of data. The
+    following wouldn't work:
 
     >>> try:
     ...     pipe(iter([1,2,3]))
@@ -435,8 +576,10 @@ def iterize(func, name=None):
 
     The solution to it is (often): ``iterize``,
     which transforms a function that is meant to be applied to a single object,
-    into a function that is meant to be applied to an array, or any iterable of such objects.
-    (You might be familiar (if you use `numpy` for example) with the related concept of "vectorization",
+    into a function that is meant to be applied to an array, or any iterable
+    of such objects.
+    (You might be familiar (if you use `numpy` for example) with the related
+    concept of "vectorization",
     or [array programming](https://en.wikipedia.org/wiki/Array_programming).)
 
 
@@ -446,10 +589,33 @@ def iterize(func, name=None):
     >>> pipe = Pipeline(iterize(lambda x: x * 2),
     ...                 iterize(lambda x: f"hello {x}"))
     >>> iterable = pipe([1, 2, 3])
-    >>> assert isinstance(iterable, Iterable)  # see that the result is an iterable
+    >>> # see that the result is an iterable
+    >>> assert isinstance(iterable, Iterable)
     >>> list(iterable)  # consume the iterable and gather it's items
     ['hello 2', 'hello 4', 'hello 6']
     """
+    # # TODO: Try replacing with partial_plus instead
+    # wrapper = mywraps(
+    #     func, name=name, doc_prefix=f"generator version of {func_name(func)}:\n"
+    # )
+    #
+    # _func = partial(map, func)
+    # new_sig = Sig(map).normalize_kind(kind=Parameter.POSITIONAL_ONLY)
+    #
+    # @wrapper
+    # @new_sig
+    # def __func(*args):
+    #     return _func(*args)
+    #
+    # __func._iterized = True
+    # return __func
+
+    # the simpler earlier version has problems with LineParametrized
+    #   TypeError: map() takes no keyword arguments
+    # because
+    #   args, kwargs = Sig(func).source_args_and_kwargs(*args, **kwargs)
+    # made kwargs that made map partial choke.
+
     wrapper = mywraps(
         func, name=name, doc_prefix=f"generator version of {func_name(func)}:\n"
     )
@@ -457,7 +623,7 @@ def iterize(func, name=None):
 
 
 def wrap_first_arg_in_list(func):
-    """Takes a func(X,...) function and returns a func([X],...) function. """
+    """Takes a func(X,...) function and returns a func([X],...) function."""
 
     @wraps(func)
     def _func(*args, **kwargs):
@@ -469,8 +635,10 @@ def wrap_first_arg_in_list(func):
 
 def deiterize(func):
     """The inverse of iterize.
-    Takes an "iterized" (a.k.a. "vectorized") function (i.e. a function that works on iterables), and
-    That is, takes a func(X,...) function and returns a next(iter(func([X],...))) function."""
+    Takes an "iterized" (a.k.a. "vectorized") function (i.e. a function that
+    works on iterables), and
+    That is, takes a func(X,...) function and returns a next(iter(func([X],
+    ...))) function."""
     return Line(wrap_first_arg_in_list(func), iter, next)
 
 
@@ -530,7 +698,8 @@ def expanded_args(func):
 
 class Enumerate:
     """Decorator a function so it enumerates the number of calls.
-    Or in general, returns (cursor, func(x)) instead of just func(x), where the start and step of the cursor can
+    Or in general, returns (cursor, func(x)) instead of just func(x),
+    where the start and step of the cursor can
     be defined (default is start=0 and step=1)
 
     >>> def foo(x):
@@ -561,7 +730,8 @@ class Enumerate:
 
 def with_cursor(func, start=0, step=1):
     """Decorator a function so it enumerates the number of calls.
-    Or in general, returns (cursor, func(x)) instead of just func(x), where the start and step of the cursor can
+    Or in general, returns (cursor, func(x)) instead of just func(x),
+    where the start and step of the cursor can
     be defined (default is start=0 and step=1)
 
     >>> def foo(x):
@@ -590,7 +760,6 @@ def with_cursor(func, start=0, step=1):
 
 
 Stats = Any
-from typing import Optional
 
 from typing import cast
 
@@ -601,9 +770,11 @@ _no_value_specified_sentinel = cast(int, object())
 
 
 class BufferStats(deque):
-    """A callable (fifo) buffer. Calls add input to it, but also returns a function of it's contents.
+    """A callable (fifo) buffer. Calls add input to it, but also returns a
+    function of it's contents.
 
-    What "add" means is configurable (through ``add_new_val`` arg). Default is append, but can be extend etc.
+    What "add" means is configurable (through ``add_new_val`` arg). Default
+    is append, but can be extend etc.
 
     >>> bs = BufferStats(maxlen=4, func=sum)
     >>> list(map(bs, range(7)))
@@ -630,10 +801,12 @@ class BufferStats(deque):
     ['a', 'ba', 'cba', 'dcba', 'edcb', 'fedc', 'gfed', 'hgfe']
 
     With ``add_new_val=deque.extend``, data can be fed in chunks.
-    In the following, also see how we use iterize to get a function that takes an iterator and returns an iterator
+    In the following, also see how we use iterize to get a function that
+    takes an iterator and returns an iterator
 
     >>> from lined import iterize
-    >>> window_stats = iterize(BufferStats(maxlen=4, func=''.join, add_new_val=deque.extend))
+    >>> window_stats = iterize(BufferStats(
+    ... maxlen=4, func=''.join, add_new_val=deque.extend))
     >>> chks = ['a', 'bc', 'def', 'gh']
     >>> for x in window_stats(chks):
     ...     print(x)
@@ -642,9 +815,12 @@ class BufferStats(deque):
     cdef
     efgh
 
-    Note: To those who might think that they can optimize this for special cases: Yes you can.
-    But SHOULD you? Is it worth the increase in complexity and reduction in flexibility?
-    See https://github.com/thorwhalen/umpyre/blob/master/misc/performance_of_rolling_window_stats.md
+    Note: To those who might think that they can optimize this for special
+    cases: Yes you can.
+    But SHOULD you? Is it worth the increase in complexity and reduction in
+    flexibility?
+    See https://github.com/thorwhalen/umpyre/blob/master/misc
+    /performance_of_rolling_window_stats.md
 
     """
 
@@ -660,10 +836,14 @@ class BufferStats(deque):
         """
 
         :param maxlen: Size of the buffer
-        :param func: The function to be computed (on buffer contents) and returned when buffer is "called"
-        :param add_new_val: The function that adds values on the buffer. Signature must be (self, new_val)
-            Is usually a deque method (``deque.append`` by default, but could be ``deque.extend``,
-            ``deque.appendleft`` etc.). Can also be any other function that has a valid (self, new_val) signature.
+        :param func: The function to be computed (on buffer contents) and
+        returned when buffer is "called"
+        :param add_new_val: The function that adds values on the buffer.
+        Signature must be (self, new_val)
+            Is usually a deque method (``deque.append`` by default, but could
+            be ``deque.extend``,
+            ``deque.appendleft`` etc.). Can also be any other function that
+            has a valid (self, new_val) signature.
         """
         if maxlen is _no_value_specified_sentinel:
             raise TypeError("You are required to specify maxlen")
@@ -673,7 +853,8 @@ class BufferStats(deque):
         super().__init__(values, maxlen=maxlen)
         self.func = func
         if isinstance(add_new_val, str):
-            add_new_val = getattr(self, add_new_val)  # add_new_val is a method of deque
+            add_new_val = getattr(self, add_new_val)  # add_new_val is a method of
+            # deque
         self.add_new_val = add_new_val
         self.__name__ = "BufferStats"
 
